#
# Copyright 2025 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import base64
import logging
import orjson
import shutil
import socket
import sys
import time
import yaml
from pathlib import Path
from flask import Flask, render_template, request, redirect
from flask_sock import Sock
from flask import cli
from ocp_vscode.comms import MessageType
from ocp_vscode.backend import ViewerBackend
from ocp_vscode.backend_logo import logo
from ocp_vscode.state import add_port, del_port
import pyperclip

CONFIG_FILE = Path.home() / ".ocpvscode_standalone"

DEFAULTS = {
    "debug": False,
    "no_glass": False,
    "no_tools": False,
    "tree_width": 240,
    "theme": "browser",
    "control": "trackball",
    "modifier_keys": {
        "shift": "shiftKey",
        "ctrl": "ctrlKey",
        "meta": "metaKey",
        "alt": "altKey",
    },
    "new_tree_behavior": True,
    "pan_speed": 0.5,
    "rotate_speed": 1.0,
    "zoom_speed": 0.5,
    "axes": False,
    "axes0": False,
    "grid_xy": False,
    "grid_xz": False,
    "grid_yz": False,
    "perspective": False,
    "transparent": False,
    "black_edges": False,
    "collapse": "R",
    "reset_camera": "RESET",
    "up": "Z",
    "ticks": 5,
    "center_grid": False,
    "grid_font_size": 12,
    "default_opacity": 0.5,
    "explode": False,
    "default_edgecolor": "#808080",
    "default_color": "#e8b024",
    "default_thickedgecolor": "MediumOrchid",
    "default_facecolor": "Violet",
    "default_vertexcolor": "MediumOrchid",
    "angular_tolerance": 0.2,
    "deviation": 0.1,
    "ambient_intensity": 1.0,
    "direct_intensity": 1.1,
    "metalness": 0.3,
    "roughness": 0.65,
}

SCRIPTS = """
    <script type="module" src="static/js/three-cad-viewer.esm.js"></script>
    <script type="module" srv="static/js/comms.js"></script>
    <script type="module" srv="static/js/logo.js"></script>
"""

JS = "./static/js/three-cad-viewer.esm.js"
CSS = "./static/css/three-cad-viewer.css"

STATIC = """
        import { Comms } from "./static/js/comms.js";
        import { logo } from "./static/js/logo.js";
"""

PORT = 0


def cleanup():
    print(f"Cleaning up with port {PORT}...")
    del_port(PORT)


atexit.register(cleanup)


def COMMS(host, port):
    return f"""
        const comms = new Comms("{host}", {port});
        const vscode = {{postMessage: (msg) => {{
                comms.sendStatus(msg);
            }}
        }};
        const standaloneViewer = () => {{
            
            const ocpLogo = logo();
            decode(ocpLogo);
            
            viewer = showViewer(ocpLogo.data.shapes, ocpLogo.config);
            window.viewer = viewer;
        }}
        window.showViewer = standaloneViewer;
    """


INIT = """onload="showViewer()" """


def save_png_data_url(data_url, output_path):
    base64_data = data_url.split(",")[1]
    image_data = base64.b64decode(base64_data)
    suffix = "-temp" + hex(int(time.time() * 1e6))[2:]
    try:
        # first write to a temp name to avoid polling is successful before finished ...
        with open(output_path + suffix, "wb") as f:
            f.write(image_data)
        # ... and then rename to the actual filename
        shutil.move(output_path + suffix, output_path)

        print(f"Wrote png file to {output_path}")
    except Exception as ex:
        print("Cannot save png file:", str(ex))


def is_port_in_use(port, host="127.0.0.1"):
    """
    Check if a port is in use on both IPv4 and IPv6 localhost.

    This function checks multiple addresses to handle dual-stack scenarios
    where a server might be listening on IPv6 and accepting IPv4 connections.

    Args:
        port: The port number to check
        host: The host to check (default: "127.0.0.1")

    Returns:
        True if the port is in use on either IPv4 or IPv6, False otherwise
    """
    import errno as errno_module
    import sys

    hosts_to_check = []

    # Determine which hosts to check based on the requested host
    if host == "127.0.0.1" or host == "localhost":
        # Check both IPv4 and IPv6 localhost
        hosts_to_check = [
            ("127.0.0.1", socket.AF_INET),
            ("::1", socket.AF_INET6),
        ]
    elif host == "0.0.0.0":
        # Check all interfaces (both IPv4 and IPv6)
        hosts_to_check = [
            ("0.0.0.0", socket.AF_INET),
            ("::", socket.AF_INET6),
        ]
    else:
        # For specific hosts, determine the address family
        try:
            info = socket.getaddrinfo(host, port, socket.AF_UNSPEC, socket.SOCK_STREAM)
            if info:
                family = info[0][0]
                hosts_to_check = [(host, family)]
        except socket.gaierror:
            # If we can't resolve the host, try IPv4 by default
            hosts_to_check = [(host, socket.AF_INET)]

    for check_host, family in hosts_to_check:
        # First try a connection test (more reliable, especially on Windows)
        try:
            test_sock = socket.socket(family, socket.SOCK_STREAM)
            test_sock.settimeout(0.5)
            result = test_sock.connect_ex((check_host, port))
            test_sock.close()

            # If connection succeeds or is refused, port is in use
            if result == 0:  # Connected successfully
                return True
            # ECONNREFUSED means nothing is listening, port is free
            # Any other error, we'll fall through to the bind test
        except (socket.error, socket.timeout):
            # Connection test failed, try bind test
            pass

        # Bind test - try to bind to the port
        try:
            bind_sock = socket.socket(family, socket.SOCK_STREAM)
            # SO_REUSEADDR allows binding to TIME_WAIT sockets
            bind_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            # For IPv6, try to set IPV6_V6ONLY to avoid dual-stack issues
            if family == socket.AF_INET6 and hasattr(socket, "IPV6_V6ONLY"):
                try:
                    bind_sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
                except (socket.error, OSError):
                    pass  # Not critical if this fails

            bind_sock.bind((check_host, port))
            bind_sock.close()
        except OSError as e:
            # Check for EADDRINUSE across platforms
            # macOS: errno.EADDRINUSE = 48
            # Linux: errno.EADDRINUSE = 98
            # Windows: errno.WSAEADDRINUSE = 10048
            if e.errno == errno_module.EADDRINUSE or (
                sys.platform == "win32" and e.errno == 10048
            ):
                return True
            # Other errors (like EADDRNOTAVAIL for IPv6 when disabled)
            # are not definitive, so continue checking
        except Exception:
            # Unexpected error - continue to check other addresses
            pass
        finally:
            try:
                bind_sock.close()
            except:
                pass

    return False


class Viewer:
    def __init__(self, params):

        self.status = {}
        self.config = {}
        self.debug = params.get("debug", False)
        self.params = params
        self.host = params.get("host", "127.0.0.1")
        self.port = params.get("port", 3939)

        self.last_changes = {}
        self.last_config = {}
        self.configure(self.params)

        if not self.debug:
            cli.show_server_banner = lambda *args, **kwargs: None
            log = logging.getLogger("werkzeug")
            log.setLevel(logging.ERROR)

        self.app = Flask(__name__)
        self.sock = Sock(self.app)

        self.backend = ViewerBackend(self.port)

        self.python_client = None
        self.javascript_client = None
        self.splash = True

        self.sock.route("/")(self.handle_message)
        self.app.add_url_rule("/viewer", "viewer", self.index)
        self.app.add_url_rule(
            "/", "redirect_to_viewer", lambda: redirect("/viewer", code=302)
        )

    def debug_print(self, *msg):
        if self.debug:
            print("Debug:", *msg)

    @staticmethod
    def _dict_diff(d1, d2):
        result = {}
        for key in d2:
            if key not in d1:
                result[key] = d2[key]
            elif d2[key] != d1[key]:
                result[key] = d2[key]
        return result

    def configure(self, params):

        # Start with defaults
        local_config = DEFAULTS.copy()

        # Then apply everything from the config file if it exists
        config_file = CONFIG_FILE
        if config_file.exists():
            with open(config_file, "r") as f:
                defaults = yaml.safe_load(f)
                for k, v in defaults.items():
                    local_config[k] = v

        local_config = dict(sorted(local_config.items()))

        # Get all params != their default value and apply it
        grid = [
            local_config["grid_xy"],
            local_config["grid_xz"],
            local_config["grid_yz"],
        ]
        for k, v in params.items():
            if k == "port":
                self.port = v
            elif k == "host":
                self.host = v
            elif k not in ["create_configfile"]:
                if v != local_config.get(k):
                    if k == "grid_xy":
                        grid[0] = True
                    elif k == "grid_xz":
                        grid[1] = True
                    elif k == "grid_yz":
                        grid[2] = True
                    else:
                        local_config[k] = v
        local_config["grid"] = grid
        local_config["reset_camera"] = local_config["reset_camera"].upper()

        local_config = dict(sorted(local_config.items()))

        for k, v in local_config.items():
            if k in ["grid_xy", "grid_xz", "grid_yz"]:
                continue
            if k == "collapse":
                self.config["collapse"] = str(v)
            elif k == "no_glass":
                self.config["glass"] = not v
            elif k == "no_tools":
                self.config["tools"] = not v
            elif k == "perspective":
                self.config["ortho"] = not v

            else:
                self.config[k] = v

        # for compatibility with 2.9.0
        if (
            self.config.get("modifier_keys") is not None
            and self.config["modifier_keys"].get("alt") is None
        ):
            self.config["modifier_keys"]["alt"] = "altKey"

        diff_config = Viewer._dict_diff(self.last_config, self.config)
        if len(diff_config) > 0:
            self.debug_print("\nConfig:", self.config)
        self.last_config = dict(self.config)

    def start(self):
        global PORT
        # Check if port is in use on both IPv4 and IPv6
        if is_port_in_use(self.port, self.host):
            print(
                f"Port {self.port} is already in use. "
                "Please choose a different port or stop the other service using this port."
            )
            sys.exit(1)

        self.backend.load_model(logo)
        add_port(self.port)
        PORT = self.port

        print(f"Info: OCP CAD Viewer runs at http://{self.host}:{self.port}")

        self.app.run(
            port=self.port, host=self.host, debug=self.debug, use_reloader=False
        )

    def index(self):
        # The browser will connect with an ip/hostname that is reachable from remote.
        # Use this ip/hostname for the websocket connection
        address, port = request.host.split(":")
        return render_template(
            "viewer.html",
            standalone_scripts=SCRIPTS,
            standalone_imports=STATIC,
            standalone_comms=COMMS(address, port),
            standalone_init=INIT,
            styleSrc=CSS,
            scriptSrc=JS,
            treeWidth=self.config["tree_width"],
            **self.config,
        )

    def not_registered(self):
        print(
            "\nNo browser registered. Please open the viewer in a browser or refresh the viewer page\n"
        )

    def handle_message(self, ws):
        while True:
            data = ws.receive()
            if isinstance(data, bytes):
                data = data.decode("utf-8")

            message_type = data[0]
            data = data[2:]

            if message_type == "C":
                self.python_client = ws
                cmd = orjson.loads(data)
                if cmd == "status":
                    self.debug_print(f"[{message_type}] Received status command")
                    self.python_client.send(orjson.dumps({"text": self.status}))

                elif cmd == "config":
                    self.debug_print(f"[{message_type}] Received config command")
                    self.configure(self.params)
                    self.config["_splash"] = self.splash
                    self.python_client.send(orjson.dumps(self.config))

                elif cmd.get("type") == "screenshot":
                    self.debug_print(f"[{message_type}] Received screenshot command")
                    if self.javascript_client is None:
                        self.not_registered()
                        continue
                    self.javascript_client.send(data)

            elif message_type == "D":
                self.python_client = ws
                self.debug_print(f"[{message_type}] Received a new model")
                if self.javascript_client is None:
                    self.not_registered()
                    continue
                self.javascript_client.send(data)
                if self.splash:
                    self.splash = False

            elif message_type == "U":
                self.javascript_client = ws
                message = orjson.loads(data)
                if message["command"] == "screenshot":
                    filename = message["text"]["filename"]
                    data_url = message["text"]["data"]
                    self.debug_print(
                        f"[{message_type}:{message['command']}] Received screenshot data for file",
                        filename,
                    )
                    save_png_data_url(data_url, filename)
                elif message["command"] == "log":
                    self.debug_print(
                        f"[{message_type}:{message['command']}]", message["text"]
                    )
                elif message["command"] == "started":
                    self.debug_print(
                        f"[{message_type}:{message['command']}] Viewer has started"
                    )
                else:
                    changes = message["text"]
                    diff_changes = Viewer._dict_diff(self.last_changes, changes)
                    if len(diff_changes) > 0:
                        self.debug_print(
                            f"[{message_type}:{message['command']}] Received incremental UI changes",
                            diff_changes,
                        )
                    self.last_changes = dict(changes)

                    for key, value in changes.items():
                        if key == "selected":
                            pyperclip.copy((",").join(changes.get("selected", [])))

                        self.status[key] = value
                    self.backend.handle_event(changes, MessageType.UPDATES)

            elif message_type == "S":
                self.python_client = ws
                self.debug_print(f"[{message_type}] Received a config")
                if self.javascript_client is None:
                    self.not_registered()
                    continue
                self.javascript_client.send(data)
                self.debug_print("Posted config to view")

            elif message_type == "L":
                self.javascript_client = ws
                print("Info: Browser as viewer client registered")

            elif message_type == "B":
                model = orjson.loads(data)["model"]
                self.backend.handle_event(model, MessageType.DATA)
                self.debug_print(f"[{message_type}] Model data sent to the backend")

            elif message_type == "R":
                self.python_client = ws
                if self.javascript_client is None:
                    self.not_registered()
                    continue
                self.javascript_client.send(data)
                self.debug_print(f"[{message_type}] Backend response received.", data)



from math import pi

from ocp_tessellate.ocp_utils import (
    area,
    axis_to_line,
    BoundingBox,
    center_of_geometry,
    center_of_mass,
    downcast,
    dist_shapes,
    get_curve,
    get_plane,
    get_point,
    get_surface,
    is_closed,
    is_topods_edge,
    is_topods_face,
    is_topods_solid,
    is_topods_compound,
    is_topods_compsolid,
    is_topods_vertex,
    is_vector,
    length,
    position_at,
    tangent_edge_at,
    rect,
    vertex,
    volume,
)


def get_shape_type(shape):
    if is_topods_vertex(shape) or is_vector(shape):
        return "Vertex"
    elif is_topods_edge(shape):
        return "Edge"
    elif is_topods_face(shape):
        return "Face"
    elif is_topods_solid(shape):
        return "Solid"
    elif is_topods_compsolid(shape):
        return "CompSolid"
    elif is_topods_compound(shape):
        return "Compound"
    else:
        print(shape, "unknown")
        return "Unknown"


def get_geom_type(shape):
    if is_topods_vertex(shape) or is_vector(shape):
        return "Point"
    elif is_topods_edge(shape):
        return get_curve(shape).GetType().name.split("_")[-1]
    elif is_topods_face(shape):
        return get_surface(shape).GetType().name.split("_")[-1]
    elif (
        is_topods_solid(shape)
        or is_topods_compound(shape)
        or is_topods_compsolid(shape)
    ):
        return "Other"
    else:
        print(shape, "unknown")
        return "Unknown"


def get_properties(shape):
    shape = downcast(shape)
    shape_type = get_shape_type(shape)
    geom_type = get_geom_type(shape)

    response = {
        "shape_type": shape_type,
        "geom_type": geom_type,
    }

    if shape_type == "Vertex":
        response["XYZ"] = get_point(shape)
        response["refpoint"] = get_point(shape)

    elif shape_type == "Edge":
        shape = downcast(shape)
        if geom_type == "Line":
            response["Start"] = get_point(position_at(shape, 0))
            response["Middle"] = get_point(position_at(shape, 0.5))
            response["End"] = get_point(position_at(shape, 1))
            response["refpoint"] = get_point(position_at(shape, 0.5))

        elif geom_type == "Circle":
            circle = get_curve(shape).Circle()
            response["Center"] = get_point(circle.Location())
            response["Radius"] = circle.Radius()
            response["Start"] = get_point(position_at(shape, 0))
            if not is_closed(shape):
                # response["geom_type"] += " (Arc)"
                response["End"] = get_point(position_at(shape, 1))
                response["refpoint"] = get_point(position_at(shape, 0.5))
            else:
                response["refpoint"] = get_point(position_at(shape, 0))

        elif geom_type == "Ellipse":
            ellipse = get_curve(shape).Ellipse()
            response["Center"] = get_point(ellipse.Location())
            response["Major radius"] = ellipse.MajorRadius()
            response["Minor radius"] = ellipse.MinorRadius()
            # response["focus1"] = get_point(ellipse.Focus1())
            # response["focus2"] = get_point(ellipse.Focus2())
            response["Start"] = get_point(position_at(shape, 0))
            if not is_closed(shape):
                # response["geom_type"] += " (Arc)"
                response["End"] = get_point(position_at(shape, 1))
                response["refpoint"] = get_point(position_at(shape, 0.5))
            else:
                response["refpoint"] = get_point(position_at(shape, 0))

        elif geom_type == "Hyperbola":
            hyperbola = get_curve(shape).Hyperbola()
            response["Start"] = get_point(position_at(shape, 0))
            response["Vertex"] = get_point(hyperbola.Location())
            response["End"] = get_point(position_at(shape, 1))
            # response["focus1"] = get_point(hyperbola.Focus1())
            # response["focus2"] = get_point(hyperbola.Focus2())
            response["refpoint"] = get_point(position_at(shape, 0.5))

        elif geom_type == "Parabola":
            parabola = get_curve(shape).Parabola()
            response["Start"] = get_point(position_at(shape, 0))
            response["Vertex"] = get_point(parabola.Location())
            response["End"] = get_point(position_at(shape, 1))
            # response["focus"] = get_point(parabola.Focus())
            response["refpoint"] = get_point(position_at(shape, 0.5))

        elif geom_type in ["Bezier", "Bspline"]:
            response["Start"] = get_point(position_at(shape, 0))
            response["Middle"] = get_point(position_at(shape, 0.5))
            response["End"] = get_point(position_at(shape, 1))
            response["refpoint"] = get_point(position_at(shape, 0.5))

        elif geom_type == "OffsetCurve":
            offset = get_curve(shape).OffsetCurve()
            response["Offset"] = offset.Offset()
            response["Start"] = get_point(position_at(shape, 0))
            if is_closed(shape):
                response["End"] = get_point(position_at(shape, 1))
            response["refpoint"] = get_point(position_at(shape, 0))

        else:
            response["Start"] = get_point(position_at(shape, 0))
            response["Middle"] = get_point(position_at(shape, 0.5))
            response["End"] = get_point(position_at(shape, 1))
            response["refpoint"] = get_point(position_at(shape, 0.5))

        response["Length"] = length(shape)

        for i in [0, 1]:
            try:
                angle = calc_angle(rect(1, 1), tangent_edge_at(shape, i))
                if angle is not None:
                    response[f"Angle@{i} to XY"] = angle["angle"]
            except Exception:
                pass

    elif shape_type == "Face":
        shape = downcast(shape)
        if geom_type == "Plane":
            plane = get_surface(shape).Plane()
            response["Center"] = get_point(plane.Location())

        elif geom_type == "Cylinder":
            cylinder = get_surface(shape).Cylinder()
            response["Center"] = get_point(cylinder.Location())
            response["Radius"] = cylinder.Radius()

        elif geom_type == "Cone":
            cone = get_surface(shape).Cone()
            response["Center"] = get_point(cone.Location())
            response["Base radius"] = cone.RefRadius()
            response["Half angle"] = cone.SemiAngle() / pi * 180

        elif geom_type == "Sphere":
            sphere = get_surface(shape).Sphere()
            response["Center"] = get_point(sphere.Location())
            response["Radius"] = sphere.Radius()

        elif geom_type == "Torus":
            torus = get_surface(shape).Torus()
            response["Center"] = get_point(torus.Location())
            response["Minor radius"] = torus.MinorRadius()
            response["Major radius"] = torus.MajorRadius()
            response["refpoint"] = get_point(torus.Location())

        elif geom_type == "Bezier":
            ...

        elif geom_type == "Bspline":
            ...

        elif geom_type == "SurfaceOfRevolution":
            revolution = get_surface(shape)
            response["Axe loc"] = get_point(revolution.AxeOfRevolution().Location())
            response["Axe dir"] = get_point(revolution.AxeOfRevolution().Direction())
            response["geom_type"] = "Revolution"

        response["Area"] = area(shape)
        response["refpoint"] = center_of_geometry(shape)

        try:
            angle = calc_angle(rect(1, 1), shape)
            if angle is not None:
                response["Angle to XY"] = angle["angle"]
        except Exception:
            pass

    elif shape_type in ["Solid", "CompSolid", "Compound"]:
        shape = downcast(shape)
        response["Volume"] = volume(shape)
        response["refpoint"] = center_of_mass(shape)

    else:
        print(f"unknown shape {type(shape)}")

    if shape_type != "Vertex":
        bb = BoundingBox(shape, optimal=True)
        response["bb"] = {
            "min": [bb.xmin, bb.ymin, bb.zmin],
            "center": bb.center,
            "max": [bb.xmax, bb.ymax, bb.zmax],
            "size": [bb.xsize, bb.ysize, bb.zsize],
        }

    return response


def get_center(shape):
    shape_type = get_shape_type(shape)
    geom_type = get_geom_type(shape)

    if shape_type == "Vertex":
        center = shape

    elif shape_type == "Edge":
        if geom_type == "Circle":
            circle = get_curve(shape).Circle()
            center = circle.Location()

        elif geom_type == "Ellipse":
            ellipse = get_curve(shape).Ellipse()
            center = ellipse.Location()

        elif geom_type == "Hyperbola":
            hyperbola = get_curve(shape).Hyperbola()
            center = hyperbola.Location()

        elif geom_type == "Parabola":
            parabola = get_curve(shape).Parabola()
            center = parabola.Location()

        else:
            center = position_at(shape, 0.5)

    elif shape_type == "Face":
        if geom_type == "Cylinder":
            cylinder = get_surface(shape).Cylinder()
            center = cylinder.Location()

        elif geom_type == "Cone":
            cone = get_surface(shape).Cone()
            center = cone.Location()

        elif geom_type == "Sphere":
            sphere = get_surface(shape).Sphere()
            center = sphere.Location()

        elif geom_type == "Torus":
            torus = get_surface(shape).Torus()
            center = torus.Location()

        elif geom_type == "SurfaceOfRevolution":
            revolution = get_surface(shape)
            center = revolution.AxeOfRevolution().Location()

        else:
            center = center_of_mass(shape)

    elif shape_type == "Solid":
        center = center_of_mass(shape)

    else:
        print(shape_type, shape)

    if is_topods_vertex(center):
        return center
    else:
        return vertex(center)


def calc_distance(shape1, shape2, center=False):
    shape1 = downcast(shape1)
    shape2 = downcast(shape2)

    if center:
        p1 = get_center(shape1)
        p2 = get_center(shape2)
        dist, p1, p2 = dist_shapes(p1, p2)
    else:
        dist, p1, p2 = dist_shapes(shape1, shape2)
    p1 = get_point(p1)
    p2 = get_point(p2)
    xdist = abs(p2[0] - p1[0])
    ydist = abs(p2[1] - p1[1])
    zdist = abs(p2[2] - p1[2])
    return {
        "distance": dist,
        "⇒ X | Y | Z": [xdist, ydist, zdist],
        "Point 1": p1,
        "Point 2": p2,
        "info": "center" if center else "min",
        "refpoint1": p1,
        "refpoint2": p2,
    }


def calc_angle(shape1, shape2):
    shape1 = downcast(shape1)
    shape2 = downcast(shape2)

    shape_type1 = get_shape_type(shape1)
    shape_type2 = get_shape_type(shape2)
    geom_type1 = get_geom_type(shape1)
    geom_type2 = get_geom_type(shape2)

    def angle_line_line(line1, line2):
        if is_topods_edge(line1):
            l1 = get_curve(line1).Line()
        else:
            l1 = line1
        if is_topods_edge(line2):
            l2 = get_curve(line2).Line()
        else:
            l2 = line2
        return l1.Angle(l2) / pi * 180

    def angle_line_plane(line, plane):
        # 90 - angle between line and normal
        axis = plane.Axis()
        line2 = axis_to_line(axis)
        return 90 - angle_line_line(line, line2)

    def angle_plane_plane(plane1, plane2):
        # angle between the two normals
        axis1 = plane1.Axis()
        line1 = axis_to_line(axis1)
        axis2 = plane2.Axis()
        line2 = axis_to_line(axis2)
        return angle_line_line(line1, line2)

    angle = None
    info1 = ""
    info2 = ""

    if geom_type1 == "Line" and geom_type2 == "Line":
        angle = angle_line_line(shape1, shape2)
        info1 = "Line"
        info2 = "Line"

    elif geom_type1 == "Line" and shape_type2 in ["Edge", "Face"]:
        plane2 = get_plane(shape2)
        if plane2 is not None:
            angle = angle_line_plane(shape1, plane2)
            info1 = "Line"
            if shape_type2 == "Edge":
                info2 = f"Plane({shape_type2})"
            else:
                info2 = f"{shape_type2}"

    elif shape_type1 in ["Edge", "Face"] and geom_type2 == "Line":
        plane1 = get_plane(shape1)
        if plane1 is not None:
            angle = angle_line_plane(shape2, plane1)
            if shape_type1 == "Edge":
                info1 = f"Plane({shape_type1})"
            else:
                info1 = f"{shape_type1}"
            info2 = "Line"

    elif shape_type1 in ["Edge", "Face"] and shape_type2 in ["Edge", "Face"]:
        plane1 = get_plane(shape1)
        plane2 = get_plane(shape2)
        if plane1 is not None and plane2 is not None:
            angle = angle_plane_plane(plane1, plane2)
            if shape_type1 == "Edge":
                info1 = f"Plane({shape_type1})"
            else:
                info1 = f"{shape_type1}"
            if shape_type2 == "Edge":
                info2 = f"Plane({shape_type2})"
            else:
                info2 = f"{shape_type2}"

    if angle is None:
        return None
    else:
        return {
            "angle": angle,
            "info1": info1,
            "info2": info2,
        }


def get_distance(shape1, shape2, center):
    response = calc_distance(shape1, shape2, center)

    angle = calc_angle(shape1, shape2)

    if angle is not None:
        response["angle"] = angle["angle"]
        response["info1"] = angle["info1"]
        response["info2"] = angle["info2"]

    return response



"""Communication with the viewer"""

#
# Copyright 2025 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import enum
import json
import os
import socket
import warnings

import questionary

from websockets.sync.client import connect

import orjson
from ocp_tessellate.utils import Timer
from ocp_tessellate.ocp_utils import (
    is_topods_shape,
    is_toploc_location,
    serialize,
    loc_to_tq,
)
from .state import get_ports, update_state, get_config_file

from IPython import get_ipython

# pylint: disable=unused-import
try:
    import jupyter_console  # noqa: F401

    JCONSOLE = True
except Exception:
    JCONSOLE = False

CMD_URL = "ws://127.0.0.1"
CMD_PORT = 3939

INIT_DONE = False


warnings.simplefilter("once", UserWarning)


def warn_once(message):
    def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
        return "%s: %s\n" % (category.__name__, message)

    warnings.formatwarning = warning_on_one_line
    warnings.warn(message, UserWarning)


#
# Send data to the viewer
#


class MessageType(enum.IntEnum):
    """Message types"""

    DATA = 1
    COMMAND = 2
    UPDATES = 3
    LISTEN = 4
    BACKEND = 5
    BACKEND_RESPONSE = 6
    CONFIG = 7


__all__ = [
    "send_data",
    "send_command",
    "send_response",
    "set_port",
    "get_port",
    "listener",
    "is_pytest",
]


def is_pytest():
    return os.environ.get("OCP_VSCODE_PYTEST") == "1"


def port_check(port):
    """Check whether the port is listening"""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(1)
    result = s.connect_ex((get_host(), port)) == 0
    if result:
        s.close()
    return result


def default(obj):
    """Default JSON serializer."""
    if is_topods_shape(obj):
        return base64.b64encode(serialize(obj)).decode("utf-8")
    elif is_toploc_location(obj):
        return loc_to_tq(obj)
    elif isinstance(obj, enum.Enum):
        return obj.value
    else:
        raise TypeError(f"Object of type {type(obj)} is not JSON serializable")


def get_port():
    """Get the port"""
    if is_pytest():
        return 3939

    if not INIT_DONE:
        find_and_set_port()
        set_connection_file()
    return CMD_PORT


def get_host():
    """Get the host"""
    return CMD_URL[5:]


def set_port(port, host="127.0.0.1"):
    """Set the port"""
    global CMD_PORT, CMD_URL, INIT_DONE  # pylint: disable=global-statement
    CMD_PORT = port
    CMD_URL = f"ws://{host}"
    INIT_DONE = True


def _send(data, message_type, port=None, timeit=False):
    """Send data to the viewer"""
    global WS

    if port is None:
        if not INIT_DONE:
            find_and_set_port()
            set_connection_file()
        port = CMD_PORT
    try:
        with Timer(timeit, "", "json dumps", 1):
            j = orjson.dumps(data, default=default)  # pylint: disable=no-member
            if message_type == MessageType.COMMAND:
                j = b"C:" + j
            elif message_type == MessageType.DATA:
                j = b"D:" + j
            elif message_type == MessageType.LISTEN:
                j = b"L:" + j
            elif message_type == MessageType.BACKEND:
                j = b"B:" + j
            elif message_type == MessageType.BACKEND_RESPONSE:
                j = b"R:" + j
            elif message_type == MessageType.CONFIG:
                j = b"S:" + j

        with Timer(timeit, "", f"websocket connect ({message_type.name})", 1):
            try:
                with connect(f"{CMD_URL}:{port}", close_timeout=0.05) as ws:
                    ws.send(j)

                    with Timer(
                        timeit, "", f"websocket send {len(j) / 1024 / 1024:.3f} MB", 1
                    ):
                        result = None
                        if message_type == MessageType.COMMAND and not (
                            isinstance(data, dict) and data["type"] == "screenshot"
                        ):
                            try:
                                result = json.loads(ws.recv())
                            except Exception as ex:  # pylint: disable=broad-except
                                print(ex)
                        elif message_type == MessageType.COMMAND and (
                            isinstance(data, dict) and data["type"] == "screenshot"
                        ):
                            result = {}

            except Exception as ex:
                warn_once("The viewer doesn't seem to run: " + str(ex))
                # set some dummy values to avoid errors
                return {
                    "collapse": "none",
                    "_splash": False,
                    "default_facecolor": (1, 234, 56),
                    "default_thickedgecolor": (123, 45, 6),
                    "default_vertexcolor": (123, 45, 6),
                }

        return result

    except Exception as ex:  # pylint: disable=broad-except
        print(
            f"Cannot connect to viewer on port {port}, is it running and the right port provided?"
        )
        print(ex)
        return None


def send_data(data, port=None, timeit=False):
    """Send data to the viewer"""
    return _send(data, MessageType.DATA, port, timeit)


def send_config(config, port=None, title=None, timeit=False):
    """Send config to the viewer"""
    return _send(config, MessageType.CONFIG, port, timeit)


def send_command(data, port=None, title=None, timeit=False):
    """Send command to the viewer"""
    result = _send(data, MessageType.COMMAND, port, timeit)
    if result.get("command") == "status":
        return result["text"]
    else:
        return result


def send_backend(data, port=None, timeit=False):
    """Send data to the viewer"""
    return _send(data, MessageType.BACKEND, port, timeit)


def send_response(data, port=None, timeit=False):
    """Send data to the viewer"""
    return _send(data, MessageType.BACKEND_RESPONSE, port, timeit)


#
# Receive data from the viewer
#


# async listener for the websocket class
# this will be called when the viewer sends data
# the data is then passed to the callback function
#
def listener(callback):
    """Listen for data from the viewer"""

    def _listen():
        last_config = {}
        with connect(f"{CMD_URL}:{CMD_PORT}", max_size=2**28) as websocket:
            websocket.send(b"L:Python listener")
            while True:
                try:
                    message = websocket.recv()
                    if message is None:
                        continue

                    message = json.loads(message)
                    if "model" in message.keys():
                        callback(message["model"], MessageType.DATA)

                    if message.get("command") == "status":
                        changes = message["text"]
                        new_changes = {}
                        for k, v in changes.items():
                            if k in last_config and last_config[k] == v:
                                continue
                            new_changes[k] = v
                        last_config = changes
                        callback(new_changes, MessageType.UPDATES)

                    elif message.get("command") == "stop":
                        print("Stopping Python listener")
                        break
                except Exception as ex:  # pylint: disable=broad-except
                    print(ex)
                    break

    return _listen


def find_and_set_port():
    """Set the port and connection file"""

    def find_port():
        port = None
        ports = get_ports()

        valid_ports = []
        for p in ports:
            if port_check(int(p)):
                valid_ports.append(p)

        if len(valid_ports) == 0:
            return None

        elif len(valid_ports) == 1:
            port = valid_ports[0]

        else:
            if get_ipython().__class__.__name__ == "ZMQInteractiveShell":
                print("\n=> Select port in VS Code input box above\n")
                port = input(f"Select port from {[int(p) for p in valid_ports]} ")
            else:
                port = questionary.select(
                    "Multiple viewers found. Select a port:",
                    choices=[str(p) for p in valid_ports],
                ).ask()
            if port is not None and port != "":
                port = int(port)

        return port

    try:
        port = int(os.environ.get("OCP_PORT", "0"))
    except ValueError:
        print(
            f"Port {os.environ.get('OCP_PORT')} taken from environment variable OCP_PORT is invalid"
        )
        port = 0

    if port > 0:
        print(f"Using predefined port {port} taken from environment variable OCP_PORT")
    else:
        port = find_port()
        if port is not None:
            print(f"Using port {port}")
        elif port_check(3939):
            port = 3939
            print(f"Default port {port} is open, using it")

    set_port(port)


def set_connection_file():
    """Set the connection file for Jupyter in the state file .ocpvscode"""
    if JCONSOLE and hasattr(get_ipython(), "kernel"):
        kernel = get_ipython().kernel
        cf = kernel.config["IPKernelApp"]["connection_file"]
        with open(cf, "r", encoding="utf-8") as f:
            connection_info = json.load(f)

        if port_check(connection_info["iopub_port"]):
            print("Jupyter kernel running")
            try:
                _ = int(CMD_PORT)
                update_state(str(CMD_PORT), cf)
                print(f"Jupyter connection file path written to {get_config_file()}")
            except ValueError:
                print(
                    f"Cannot set Jupyter connection file, port {CMD_PORT}' is non-numeric"
                )
        else:
            print("Jupyter kernel not responding")



"""Python backend for the OCP Viewer"""

#
# Copyright 2025 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import base64
from dataclasses import dataclass
import sys
import traceback
import os

from ocp_tessellate.ocp_utils import (
    deserialize,
    make_compound,
    tq_to_loc,
    identity_location,
    downcast,
)

from ocp_tessellate.tessellator import (
    get_edges,
    get_faces,
    get_vertices,
)
from ocp_tessellate.trace import Trace

from ocp_vscode.backend_logo import logo

if os.environ.get("JUPYTER_CADQUERY") is None:
    is_jupyter_cadquery = False
    from ocp_vscode.comms import send_response
else:
    is_jupyter_cadquery = True

from ocp_vscode.comms import MessageType, listener, set_port
from ocp_vscode.measure import get_distance, get_properties


@dataclass
class Tool:
    """The tools available in the viewer"""

    Distance = "DistanceMeasurement"
    Properties = "PropertiesMeasurement"


def print_to_stdout(*msg):
    """
    Write the given message to the stdout
    """
    print(*msg, flush=True, file=sys.stdout)


def error_handler(func):
    """Decorator for error handling"""

    def wrapper(*args, **kwargs):  # pylint: disable=redefined-outer-name
        try:
            return func(*args, **kwargs)
        except Exception as exc:  # pylint: disable=broad-except
            print_to_stdout(exc)
            traceback.print_exception(*sys.exc_info(), file=sys.stdout)

    return wrapper


class ViewerBackend:
    """
    Represents the backend of the viewer, it listens to the websocket and handles the events
    It's job is to send responses to the vscode extension that goes through the three cad
    viewer view.
    The responses holds all the data needed to display the measurements.
    """

    def __init__(self, port: int, jcv_id=None) -> None:
        self.port = port
        self.model = None
        self.activated_tool = None
        self.filter_type = "none"  # The current active selection filter
        self.jcv_id = jcv_id
        set_port(port)

    def start(self):
        """
        Start the backend
        """
        print("Viewer backend started")
        self.load_model(logo)
        print("Logo model loaded")
        listener(self.handle_event)()

    @error_handler
    def handle_event(self, message, event_type: MessageType):
        """
        Handle the event received from the websocket
        Dispatch the event to the appropriate handler
        """
        if event_type == MessageType.DATA:
            self.load_model(message)
        elif event_type == MessageType.UPDATES:
            changes = message

            if "activeTool" in changes:
                active_tool = changes.get("activeTool")

                if active_tool != "None":
                    self.activated_tool = active_tool
                else:
                    self.activated_tool = None

            if self.activated_tool is not None:
                return self.handle_activated_tool(changes)

    def handle_activated_tool(self, changes):
        """
        Handle the activated tool, there is a special behavior for each tool
        """
        if "selectedShapeIDs" not in changes:
            return

        selected_objs = changes["selectedShapeIDs"]
        if self.activated_tool == Tool.Distance and len(selected_objs) == 3:
            shape_id1 = changes["selectedShapeIDs"][0]
            shape_id2 = changes["selectedShapeIDs"][1]
            shift = changes["selectedShapeIDs"][2]

            return self.handle_distance(shape_id1, shape_id2, shift)

        elif self.activated_tool == Tool.Properties and len(selected_objs) == 2:
            shape_id = changes["selectedShapeIDs"][0]
            return self.handle_properties(shape_id)

    def load_model(self, raw_model):
        """Read the transferred model from websocket"""

        def walk(model, trace):
            for v in model["parts"]:
                if v.get("parts") is not None:
                    walk(v, trace)
                else:
                    id_ = v["id"]
                    loc = (
                        identity_location()
                        if v["loc"] is None
                        else tq_to_loc(*v["loc"])
                    )
                    if isinstance(v["shape"], dict):
                        compound = deserialize(
                            base64.b64decode(v["shape"]["obj"].encode("utf-8"))
                        )
                    else:
                        shape = [
                            deserialize(base64.b64decode(s.encode("utf-8")))
                            for s in v["shape"]
                        ]
                        compound = make_compound(shape) if len(shape) > 1 else shape[0]
                    self.model[id_] = compound.Moved(loc)
                    faces = get_faces(compound)
                    for i, face in enumerate(faces):
                        trace.face(f"{id_}/faces/faces_{i}", face)

                        self.model[f"{id_}/faces/faces_{i}"] = downcast(face.Moved(loc))
                    edges = get_edges(compound)
                    for i, edge in enumerate(edges):
                        trace.edge(f"{id_}/edges/edges_{i}", edge)

                        self.model[f"{id_}/edges/edges_{i}"] = (
                            edge if loc is None else downcast(edge.Moved(loc))
                        )
                    vertices = get_vertices(compound)
                    for i, vertex in enumerate(vertices):
                        trace.vertex(f"{id_}/vertices/vertex_{i}", vertex)

                        self.model[f"{id_}/vertices/vertices_{i}"] = (
                            vertex if loc is None else downcast(vertex.Moved(loc))
                        )

        self.model = {}
        trace = Trace("ocp-vscode-backend.log")
        walk(raw_model, trace)
        trace.close()

    def handle_properties(self, shape_id):
        """
        Request the properties of the object with the given id
        """
        if not is_jupyter_cadquery:
            print_to_stdout(f"Identifier received '{shape_id}'")

        shape = self.model[shape_id]

        response = get_properties(shape)

        response["type"] = "backend_response"
        response["subtype"] = "tool_response"
        response["tool_type"] = Tool.Properties

        if is_jupyter_cadquery:
            return response
        else:
            send_response(response, self.port)
            print_to_stdout(f"Data sent {response}")

    def handle_distance(self, id1, id2, center):
        """
        Request the distance between the two objects that have the given ids
        """
        if not is_jupyter_cadquery:
            print_to_stdout(f"Identifiers received '{id1}', '{id2}'")

        shape1 = self.model[id1]
        shape2 = self.model[id2]

        response = get_distance(shape1, shape2, center)

        response["type"] = "backend_response"
        response["subtype"] = "tool_response"
        response["tool_type"] = Tool.Distance

        if is_jupyter_cadquery:
            return response
        else:
            send_response(response, self.port)
            print_to_stdout(f"Data sent {response}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser("OCP Viewer Backend")
    parser.add_argument(
        "--port", type=int, required=True, help="Port the viewer listens to"
    )
    args = parser.parse_args()
    backend = ViewerBackend(args.port)

    try:
        backend.start()
    except Exception as ex:  # pylint: disable=broad-except
        print(ex)
